@@ -1,4 +1,4 @@
-/* $Header: /local/src/CVS/nickle/builtin-file.c,v 1.17 2004/06/08 09:30:53 keithp Exp $ */
+/* $Header: /local/src/CVS/nickle/builtin-file.c,v 1.19 2004/08/23 00:42:06 keithp Exp $ */
 
 /*
  * Copyright © 1988-2004 Keith Packard and Bart Massey.
@@ -215,6 +215,12 @@ do_File_print (Value file, Value value, Value format,
     ibase = IntPart (base, "Illegal base");
     if (aborting)
 	return Void;
+    if (ibase < 0 || ibase == 1)
+    {
+	RaiseStandardException (exception_invalid_argument,
+				"Illegal base", 2, NewInt (0), base);
+	return Void;
+    }
     iwidth = IntPart (width, "Illegal width");
     if (aborting)
 	return Void;
