@@ -1,4 +1,4 @@
-/* $Header: /local/src/CVS/nickle/builtin-toplevel.c,v 1.26 2004/06/08 09:30:53 keithp Exp $ */
+/* $Header: /local/src/CVS/nickle/builtin-toplevel.c,v 1.28 2004/11/23 05:02:07 bart Exp $ */
 
 /*
  * Copyright © 1988-2004 Keith Packard and Bart Massey.
@@ -130,6 +130,10 @@ import_Toplevel_namespace()
 	    " bool is_uninit (*poly r)\n"
 	    "\n"
 	    " Return whether 'r' references uninitialized storage.\n" },
+        { do_make_uninit, "make_uninit", "v", "*p", "\n"
+	    " void make_uninit (*poly r)\n"
+	    "\n"
+	    " Makes 'r' reference uninitialized storage.\n" },
         { do_mantissa, "mantissa", "r", "R", "\n"
 	    " rational mantissa (real r)\n"
 	    "\n"
@@ -446,9 +450,9 @@ do_dims(Value av)
     int ndim = av->array.ndim;
 
     ret = NewArray(True, False, typePrim[rep_int], 1, &ndim);
-    for (i = 0; i < av->array.ndim; i++) {
+    for (i = 0; i < ndim; i++) {
 	Value d = NewInt(ArrayLimits(&av->array)[i]);
-	ArrayValueSet(&ret->array, i, d);
+	ArrayValueSet(&ret->array, ndim - i - 1, d);
     }
     RETURN (ret);
 }
@@ -833,6 +837,20 @@ do_is_uninit (Value av)
     RETURN (av);
 }
 
+Value
+do_make_uninit (Value av)
+{
+    ENTER ();
+    if (!av) {
+	RaiseStandardException (exception_invalid_argument,
+				"do_make_uninit: invalid reference",
+				2, NewInt (0), av);
+    } else {
+	RefValueSet(av, 0);
+    }
+    RETURN (Void);
+}
+
 Value
 do_is_array (Value av)
 {
