@@ -1,4 +1,4 @@
-/* $Header: /local/src/CVS/nickle/builtin.c,v 1.20 2004/06/14 06:43:33 keithp Exp $ */
+/* $Header: /local/src/CVS/nickle/builtin.c,v 1.23 2004/12/17 02:05:20 keithp Exp $ */
 
 /*
  * Copyright © 1988-2004 Keith Packard and Bart Massey.
@@ -14,27 +14,6 @@
 
 #include	"builtin.h"
 
-struct sbuiltin {
-    char	    *value;
-    char	    *name;
-    NamespacePtr    *namespace;
-};
-
-struct envbuiltin {
-#ifdef CENVIRON
-    char	    *var;
-#endif
-    char	    *def;
-    char	    *name;
-    NamespacePtr    *namespace;
-};
-    
-struct filebuiltin {
-    char	    *name;
-    Value   	    *value;
-    NamespacePtr    *namespace;
-};
-
 static const struct sbuiltin svars[] = {
     { "> ",	    "prompt" },
     { "+ ",	    "prompt2" },
@@ -141,6 +120,14 @@ BuiltinSymbol (NamespacePtr *namespacep,
 					     type)));
 }
 
+static Type *typeUserdef[100];
+
+void
+BuiltinSetUserdefType (Type *type, int n)
+{
+    typeUserdef[n] = type;
+}
+
 static char *
 BuiltinType (char *format, Type **type)
 {
@@ -151,6 +138,8 @@ BuiltinType (char *format, Type **type)
     Bool    resizable = False;
     Expr    *dims = 0;
     Type    *k;
+    char    f;
+    int	    i;
     
     ref = False;
     if (*format == '*')
@@ -175,7 +164,7 @@ BuiltinType (char *format, Type **type)
 	hash = True;
 	format = BuiltinType (format + 1, &k);
     }
-    switch (*format++) {
+    switch (f = *format++) {
     case 'p': t = typePoly; break;
     case 'n': t = typePrim[rep_float]; break;
     case 'N': t = typePrim[rep_float]; break;
@@ -190,7 +179,14 @@ BuiltinType (char *format, Type **type)
     case 'c': t = typePrim[rep_continuation]; break;
     case 'b': t = typePrim[rep_bool]; break;
     case 'v': t = typePrim[rep_void]; break;
+    case 'F': t = typePrim[rep_foreign]; break;
     case 'a': t = typeSockaddr; break;
+    case '0': case '1': case '2': case '3': case '4':
+    case '5': case '6': case '7': case '8': case '9':
+	i = (f - '0') * 10 + (*format++) - '0';
+	t = typeUserdef[i];
+	if (t)
+	    break;
     default: 
 	t = 0;
 	write (2, "Invalid builtin argument type\n", 30);
@@ -306,7 +302,6 @@ void
 BuiltinInit (void)
 {
     ENTER ();
-    const struct sbuiltin	*s;
     const struct filebuiltin	*f;
     const struct ebuiltin	*e;
     SymbolPtr			sym;
@@ -333,12 +328,10 @@ BuiltinInit (void)
 #endif
     import_Environ_namespace();
     import_Socket_namespace();
+    import_Foreign_namespace ();
 
     /* Import builtin strings with predefined values */
-    for (s = svars; s->name; s++) {
-	sym = BuiltinSymbol (s->namespace, s->name, typePrim[rep_string]);
-	BoxValueSet (sym->global.value, 0, NewStrString (s->value));
-    }
+    BuiltinStrings (svars);
 
 #ifdef CENVIRON
     /* Get the user's home directory in case it's referenced in the
