@@ -1,4 +1,4 @@
-/* $Header: /local/src/CVS/nickle/execute.c,v 1.90 2004/04/16 21:41:52 keithp Exp $ */
+/* $Header: /local/src/CVS/nickle/execute.c,v 1.94 2004/12/11 06:02:23 keithp Exp $ */
 
 /*
  * Copyright © 1988-2004 Keith Packard and Bart Massey.
@@ -224,13 +224,16 @@ ThreadCall (Value thread, Bool tail, InstPtr *next, int *stack)
 		value = Void;
 		break;
 	    }
-	}
-	if (tail && !aborting)
-	{
-	    complete = True;
-	    thread->thread.continuation.obj = thread->thread.continuation.frame->saveObj;
-	    *next = thread->thread.continuation.frame->savePc;
-	    thread->thread.continuation.frame = thread->thread.continuation.frame->previous;
+	    /*
+	     * For a tail call, drop the topmost frame
+	     */
+	    if (tail && !aborting)
+	    {
+		complete = True;
+		thread->thread.continuation.obj = thread->thread.continuation.frame->saveObj;
+		*next = thread->thread.continuation.frame->savePc;
+		thread->thread.continuation.frame = thread->thread.continuation.frame->previous;
+	    }
 	}
     }
     else
@@ -321,7 +324,12 @@ ThreadArray (Value thread, Bool resizable, int ndim, Type *type)
     dims = AllocateTemp (ndim * sizeof (int));
     for (i = 0; i < ndim; i++)
     {
-	dims[i] = IntPart (Stack(i), "Invalid array dimension");
+	Value	d = Stack(i);
+	dims[i] = IntPart (d, "Invalid array dimension");
+	if (dims[i] < 0)
+	    RaiseStandardException (exception_invalid_argument,
+				    "Negative array dimension", 
+				    2, NewInt (0), d);
 	if (aborting)
 	    RETURN (0);
     }
@@ -340,7 +348,12 @@ ThreadArrayInd (Value thread, Bool resizable, Value dim, Type *type)
     dims = AllocateTemp (ndim * sizeof (int));
     for (i = 0; i < ndim; i++)
     {
-	dims[i] = IntPart (ArrayValueGet(a, i), "Invalid array dimension");
+	Value	d = ArrayValueGet(a, i);
+	dims[i] = IntPart (d, "Invalid array dimension");
+	if (dims[i] < 0)
+	    RaiseStandardException (exception_invalid_argument,
+				    "Negative array dimension", 
+				    2, NewInt (0), d);
 	if (aborting)
 	    RETURN (0);
     }
@@ -578,6 +591,23 @@ ThreadArrayInit (Value thread, Value value, AInitMode mode,
 }
 
 
+#ifdef DEBUG_JUMP
+void
+ThreadCatches (Value thread)
+{
+    CatchPtr	catch;
+
+    FilePrintf (FileStdout, "(");
+    for (catch = thread->thread.continuation.catches; 
+	 catch;
+	 catch = catch->continuation.catches)
+    {
+	FilePrintf (FileStdout, "%A ", catch->exception->symbol.name);
+    }
+    FilePrintf (FileStdout, ")\n");
+}
+#endif
+
 static Value
 ThreadRaise (Value thread, Value value, int argc, SymbolPtr exception, InstPtr *next)
 {
@@ -588,7 +618,8 @@ ThreadRaise (Value thread, Value value, int argc, SymbolPtr exception, InstPtr *
     int		base = argc - 2;
 
 #ifdef DEBUG_JUMP
-    FilePrintf (FileStdout, "    Raise: %A\n", exception->symbol.name);
+    FilePrintf (FileStdout, "    Raise: %A ", exception->symbol.name);
+    ThreadCatches (thread);
 #endif
     /*
      * Build and array to hold the arguments, this will end up
@@ -606,6 +637,20 @@ ThreadRaise (Value thread, Value value, int argc, SymbolPtr exception, InstPtr *
     RETURN (args);
 }
 
+static void
+ThreadEndCatch (Value thread, int catches)
+{
+#ifdef DEBUG_JUMP
+    FilePrintf (FileStdout, "    EndCatch: %d ", catches);
+#endif
+    while (catches--)
+        thread->thread.continuation.catches = thread->thread.continuation.catches->continuation.catches;
+#ifdef DEBUG_JUMP
+    ThreadCatches (thread);
+#endif
+}
+
+
 static Value
 ThreadExceptionCall (Value thread, InstPtr *next, int *stack)
 {
@@ -681,8 +726,22 @@ ThreadUnwind (Value thread, int twixt, int catch)
 {
     while (twixt--)
 	thread->thread.continuation.twixts = thread->thread.continuation.twixts->continuation.twixts;
+#ifdef DEBUG_JUMP
+    FilePrintf (FileStdout, "     Before unwind: ");
+    ThreadCatches (thread);
+#endif
     while (catch--)
+    {
+#ifdef DEBUG_JUMP
+	FilePrintf (FileStdout, "    Unwind: %A\n",
+			thread->thread.continuation.catches->exception->symbol.name);
+#endif
 	thread->thread.continuation.catches = thread->thread.continuation.catches->continuation.catches;
+    }
+#ifdef DEBUG_JUMP
+    FilePrintf (FileStdout, "     After unwind: ");
+    ThreadCatches (thread);
+#endif
 }
 
 #define ThreadBoxCheck(box,i) (BoxValueGet(box,i) == 0 ? ThreadBoxSetDefault(box,i,0) : 0)
@@ -884,6 +943,10 @@ ThreadStackDump (Value thread)
     }
 }
 
+#ifdef DEBUG_INST
+int dump_inst = 0;
+#endif
+
 void
 ThreadsRun (Value thread, Value lex)
 {
@@ -925,15 +988,7 @@ ThreadsRun (Value thread, Value lex)
 	else if (thread && !Runnable (thread))
 	    break;
 	else if (!running)
-	{
-	    sigset_t	set, oset;
-
-	    sigfillset (&set);
-	    sigprocmask (SIG_SETMASK, &set, &oset);
-	    if (!signaling)
-		sigsuspend (&oset);
-	    sigprocmask (SIG_SETMASK, &oset, &set);
-	}
+	    ThreadsBlock ();
 	else 
 	{
 	    ENTER ();
@@ -955,8 +1010,11 @@ ThreadsRun (Value thread, Value lex)
 		next = inst + 1;
 		complete = False;
 #ifdef DEBUG_INST
-		InstDump (inst, 1, 0, 0, 0);
-		FileFlush (FileStdout, True);
+		if (dump_inst)
+		{
+		    InstDump (inst, 1, 0, 0, 0);
+		    FileFlush (FileStdout, True);
+		}
 #endif
 		switch (inst->base.opCode) {
 		case OpNoop:
@@ -1297,6 +1355,11 @@ ThreadsRun (Value thread, Value lex)
 		case OpCatch:
 		    if (aborting)
 			break;
+#ifdef DEBUG_JUMP
+		    FilePrintf (FileStdout, "    Catch: %A ",
+				inst->catch.exception->symbol.name);
+		    ThreadCatches (thread);
+#endif
 		    thread->thread.continuation.catches = NewCatch (thread,
 								    inst->catch.exception);
 		    complete = True;
@@ -1305,7 +1368,7 @@ ThreadsRun (Value thread, Value lex)
 		case OpEndCatch:
 		    if (aborting)
 			break;
-		    thread->thread.continuation.catches = thread->thread.continuation.catches->continuation.catches;
+		    ThreadEndCatch (thread, inst->ints.value);
 		    complete = True;
 		    break;
 		case OpRaise:
