@@ -1,4 +1,4 @@
-/* $Header: /local/src/CVS/nickle/float.c,v 1.23 2004/02/27 03:50:16 keithp Exp $ */
+/* $Header: /local/src/CVS/nickle/float.c,v 1.27 2004/12/17 05:49:33 keithp Exp $ */
 
 /*
  * Copyright © 1988-2004 Keith Packard and Bart Massey.
@@ -786,8 +786,12 @@ FloatPrint (Value f, Value fv, char format, int base, int width, int prec, int f
     DebugF ("fratio ", &fratio->floats);
     negative = a->mant->sign == Negative;
     m = NewInteger (Positive, a->mant->mag);
-    if (FpartLength (a->mant) == a->prec)
-	m = Plus (m, One);
+    
+    /*
+     * Round the mantissa up by adding a bit at the extreme of the precision
+     */
+    m = Plus (m, NewFloat (one_fpart,
+			   NewIntFpart (length - a->prec), a->prec + 2));
     m = Times (m, fratio);
     if (True (Less (m, One)))
     {
@@ -910,12 +914,13 @@ FloatPrint (Value f, Value fv, char format, int base, int width, int prec, int f
 	frac_width = 0;
     frac_buffer = 0;
     frac_string = 0;
+    if (frac_width)
+	frac_part = Floor (Times (frac_part, Pow (NewInt (base), 
+						  NewInt (frac_width - 1))));
     if (frac_width && (!Zerop (frac_part) || orig_prec > 0))
     {
 	int	frac_wrote;
 	
-	frac_part = Floor (Times (frac_part, Pow (NewInt (base), 
-						  NewInt (frac_width - 1))));
 	if (ValueIsInteger(frac_part))
 	    frac_n = IntegerMag(frac_part);
 	else
@@ -936,6 +941,9 @@ FloatPrint (Value f, Value fv, char format, int base, int width, int prec, int f
 	    frac_wrote++;
 	}
 	*--frac_string = '.';
+	if (orig_prec < 0)
+	    while (frac_buffer[frac_width-1] == '0')
+		frac_buffer[--frac_width] = '\0';
     }
     else
 	frac_width = 0;
@@ -971,6 +979,15 @@ FloatPrint (Value f, Value fv, char format, int base, int width, int prec, int f
     return True;
 }
 
+static HashValue
+FloatHash (Value av)
+{
+    Float   *a = &av->floats;
+
+    return (NaturalHash(a->mant->mag) ^ a->mant->sign ^
+	    NaturalHash(a->exp->mag) ^ a->exp->sign);
+}
+
 static void
 FloatMark (void *object)
 {
@@ -1002,7 +1019,9 @@ ValueRep   FloatRep = {
     },
     FloatPromote,
     FloatReduce,
-    FloatPrint
+    FloatPrint,
+    0,
+    FloatHash,
 };
 
 Value
@@ -1083,6 +1102,33 @@ NewRationalFloat (Rational *r, unsigned prec)
     RETURN (FloatDivide (num, den, 1));
 }
 
+#define SCALE_BITS  52
+#define SCALE	    4503599627370496.0	/* 2 ** 52 */
+
+Value
+NewDoubleFloat (double d)
+{
+    ENTER ();
+    int	    e;
+    double  m;
+    Sign    ms;
+
+    double_digit    dd;
+    if (d == 0.0) RETURN (Zero);
+    e = ilogb (d);
+    m = significand (d);
+    ms = Positive;
+    if (m < 0)
+    {
+	ms = Negative;
+	m = -m;
+    }
+    e = e - SCALE_BITS;
+    dd = (double_digit) (m * SCALE + 0.5);
+    RETURN (NewFloat (NewFpart (ms, NewDoubleDigitNatural (dd)),
+		      NewIntFpart (e), SCALE_BITS));
+}
+
 Value
 NewValueFloat (Value av, unsigned prec)
 {
@@ -1106,3 +1152,41 @@ NewValueFloat (Value av, unsigned prec)
     }
     RETURN (av);
 }
+
+double
+DoublePart (Value av, char *error)
+{
+    double  mantissa;
+    int	    i;
+    int	    e;
+    digit   *mt;
+    double  div;
+    
+    av = NewValueFloat (av, 64);
+    if (NaturalLess (av->floats.exp->mag, max_int_natural))
+	e = NaturalToInt (av->floats.exp->mag);
+    else
+	e = MAX_NICKLE_INT;
+    if (e > 1023)
+    {
+	RaiseStandardException (exception_invalid_argument, error,
+				2, NewInt (0), av);
+	return 0.0;
+    }
+    if (av->floats.exp->sign == Negative)
+	e = -e;
+    
+    mantissa = 0.0;
+    i = av->floats.mant->mag->length;
+    e += DIGITBITS * i;
+    mt = NaturalDigits (av->floats.mant->mag) + i;
+    div = 1.0 / (double) BASE;
+    while (i--)
+    {
+	mantissa = mantissa + (double) *--mt * div;
+	div *= 1.0 / (double) BASE;
+    }
+    if (av->floats.mant->sign == Negative)
+	mantissa = -mantissa;
+    return mantissa * pow (2.0, (double) e);
+}
