@@ -1,4 +1,4 @@
-/* $Header: /local/src/CVS/nickle/sched.c,v 1.55 2004/04/16 05:26:45 keithp Exp $ */
+/* $Header: /local/src/CVS/nickle/sched.c,v 1.59 2004/12/11 06:02:23 keithp Exp $ */
 
 /*
  * Copyright © 1988-2004 Keith Packard and Bart Massey.
@@ -218,7 +218,7 @@ do_Thread_list (void)
 	FilePrintf (FileStdout, "\t%%%d", t->thread.id);
 	ThreadListState (t);
 	if (t->thread.sleep)
-	    FilePrintf (FileStdout, " %v", t->thread.sleep);
+	    FilePrintf (FileStdout, " %g", t->thread.sleep);
 	FileOutput (FileStdout, '\n');
     }
     RETURN(Void);
@@ -362,7 +362,7 @@ TraceFunction (Value file, FramePtr frame, CodePtr code, ExprPtr name)
     {
 	if (fe)
 	    FilePuts (file, ", ");
-	FilePrintf (file, "%v", BoxValue (frame->frame, fe));
+	FilePrintf (file, "%g", BoxValue (frame->frame, fe));
     }
     FilePuts (file, ")\n");
 }
@@ -381,7 +381,8 @@ TraceFrame (Value file, FramePtr frame, ObjPtr obj, InstPtr pc, int depth)
     int		max;
     CodePtr	code;
 
-    TraceStatement (file, ObjStatement (obj, pc));
+    if (obj && pc)
+	TraceStatement (file, ObjStatement (obj, pc));
     for (max = depth; frame && max--; frame = frame->previous)
     {
 	code = frame->function->func.code;
@@ -396,7 +397,7 @@ static void
 TraceIndent (int indent)
 {
     while (indent--)
-	FilePuts (file, "    ");
+	FilePuts (FileStdout, "    ");
 }
 #endif
 
@@ -516,7 +517,79 @@ NewThread (FramePtr frame, ObjPtr code)
     RETURN (ret);
 }
 
+typedef struct _blockHandler {
+    DataType		    *data;
+    struct _blockHandler    *next;
+    NickleBlockHandler	    handler;
+    void		    *closure;
+} BlockHandler;
+
+static void
+BlockHandlerMark (void *object)
+{
+    BlockHandler    *bh = object;
+
+    MemReference (bh->next);
+}
+
+DataType BlockHandlerType = { BlockHandlerMark, 0, "BlockHandlerType" };
+
+static BlockHandler	*blockHandlers;
+
+void
+ThreadsRegisterBlockHandler (NickleBlockHandler handler, void *closure)
+{
+    ENTER ();
+    BlockHandler    *bh = ALLOCATE (&BlockHandlerType, sizeof (BlockHandler));
+    bh->next = blockHandlers;
+    blockHandlers = bh;
+    bh->handler = handler;
+    bh->closure = closure;
+    EXIT ();
+}
+
+void
+ThreadsUnregisterBlockHandler (NickleBlockHandler handler, void *closure)
+{
+    ENTER ();
+    BlockHandler    **prev, *bh;
+
+    for (prev = &blockHandlers; (bh = *prev); prev = &bh->next)
+    {
+	if (bh->handler == handler && bh->closure == closure)
+	{
+	    bh->handler = 0;
+	    *prev = bh->next;
+	}
+    }
+    EXIT ();
+}
+
+void
+ThreadsBlock (void)
+{
+    BlockHandler    *bh, *next;
+
+    for (bh = blockHandlers; bh; bh = next)
+    {
+	next = bh->next;
+	if (bh->handler)
+	    (*bh->handler) (bh->closure);
+    }
+    if (!running)
+    {
+	sigset_t	    set, oset;
+
+	sigfillset (&set);
+	sigprocmask (SIG_SETMASK, &set, &oset);
+	if (!signaling)
+	    sigsuspend (&oset);
+	sigprocmask (SIG_SETMASK, &oset, &set);
+    }
+}
+
 ReferencePtr	RunningReference, StoppedReference;
+ReferencePtr	BlockHandlerReference;
 
 void
 ThreadInit (void)
@@ -526,6 +599,8 @@ ThreadInit (void)
     MemAddRoot (RunningReference);
     StoppedReference = NewReference ((void **) &stopped);
     MemAddRoot (StoppedReference);
+    BlockHandlerReference = NewReference ((void **) &blockHandlers);
+    MemAddRoot (BlockHandlerReference);
     EXIT ();
 }
 
@@ -573,11 +648,19 @@ FarJumpContinuation (ContinuationPtr continuation, FarJumpPtr farJump)
     /*
      * Unwind catches
      */
+#ifdef DEBUG_JUMP
+    FilePrintf (FileStdout, "FarJump catches before: ");
+    ThreadCatches (running);
+#endif
     catches = farJump->catch;
     catch = ret->continuation.catches;
     while (catches--)
 	catch = catch->continuation.catches;
     ret->continuation.catches = catch;
+#ifdef DEBUG_JUMP
+    FilePrintf (FileStdout, "FarJump catches after: ");
+    ThreadCatches (running);
+#endif
 
     /*
      * Unwind frames
@@ -685,6 +768,7 @@ EmptyContinuation (void)
 }
 
 #ifdef DEBUG_JUMP
+int dump_jump = 0;
 
 void
 ContinuationTrace (char *where, Continuation *continuation, int indent)
@@ -696,6 +780,8 @@ ContinuationTrace (char *where, Continuation *continuation, int indent)
     ObjPtr	obj = continuation->obj;
     InstPtr	pc = continuation->pc;
     
+    if (!dump_jump)
+	return;
     TraceIndent (indent);
     FilePuts (FileStdout, "*** ");
     FilePuts (FileStdout, where);
@@ -706,12 +792,16 @@ ContinuationTrace (char *where, Continuation *continuation, int indent)
     {
 	if (s)
 	    FilePuts (FileStdout, ", ");
-	FilePrintf (FileStdout, "%v", STACK_ELT(stack, s));
+	FilePrintf (FileStdout, "%g", STACK_ELT(stack, s));
     }
     FilePuts (FileStdout, "\n");
     TraceIndent (indent);
+    FilePuts (FileStdout, "frame:\nCALLS\n");
+    TraceFrame (FileStdout, continuation->frame, obj, pc, 20);
+    FilePuts (FileStdout, "END CALLS\n");
+    TraceIndent (indent);
     FilePuts (FileStdout, "catches:   ");
-    for (s = 0; catches; catches = catches->previous, s++)
+    for (s = 0; catches; catches = catches->continuation.catches, s++)
     {
 	if (s)
 	    FilePuts (FileStdout, ", ");
@@ -720,7 +810,10 @@ ContinuationTrace (char *where, Continuation *continuation, int indent)
     FilePuts (FileStdout, "\n");
     TraceIndent (indent);
     FilePuts (FileStdout, "statement: ");
-    PrettyStat (FileStdout, ObjStatement (obj, pc), False);
+    if (obj && pc)
+	PrettyStat (FileStdout, ObjStatement (obj, pc), False);
+    else
+	FilePuts (FileStdout, "corrupted continuation!\n");
     for (s = 0; twixts; twixts = twixts->continuation.twixts, s++)
     {
 	ContinuationTrace ("twixt", &twixts->continuation, indent+1);
@@ -1057,6 +1150,11 @@ RaiseException (Value thread, SymbolPtr except, Value args, InstPtr *next)
 	if (catch->exception == except)
 	{
 	    continuation = &catch->continuation;
+	    /*
+	     * Hold a reference to this nested value because
+	     * ContinuationJump is about to smash the thread
+	     */
+	    REFERENCE (catch);
 	    break;
 	}
     }
@@ -1072,7 +1170,7 @@ RaiseException (Value thread, SymbolPtr except, Value args, InstPtr *next)
 	    int	    dim = ArrayLimits(&args->array)[0];
 	    for (i = 0; i < dim; i++)
 	    {
-		PrintError ("%v", ArrayValueGet (&args->array, i));
+		PrintError ("%g", ArrayValueGet (&args->array, i));
 		if (i < dim - 1)
 		    PrintError (", ");
 	    }
