@@ -1,4 +1,4 @@
-/* $Header: /local/src/CVS/nickle/value.h,v 1.106 2004/06/14 06:43:33 keithp Exp $ */
+/* $Header: /local/src/CVS/nickle/value.h,v 1.111 2004/12/23 22:07:16 keithp Exp $ */
 
 /*
  * Copyright © 1988-2004 Keith Packard and Bart Massey.
@@ -22,8 +22,6 @@
 #include	<string.h>
 #include	<signal.h>
 #include	<assert.h>
-#include	"mem.h"
-#include	"opcode.h"
 
 typedef enum { False = 0, True = 1 }  	Bool;
 typedef char		*Atom;
@@ -267,7 +265,9 @@ typedef enum _unaryOp {
  * with functions that operate on the value
  */
 typedef enum _rep {
+	/* unknown type */
 	rep_undef = -1,
+	/* primitive types */
  	rep_int = 0,
 	rep_integer = 1,
  	rep_rational = 2,
@@ -278,15 +278,18 @@ typedef enum _rep {
 	rep_semaphore = 7,
 	rep_continuation = 8,
 	rep_bool = 9,
-	rep_void = 10,
-	rep_ref = 11,
-	rep_func = 12,
+	rep_foreign = 10,
+	rep_void = 11,
+	
+	/* composite types */
+	rep_ref = 12,
+	rep_func = 13,
     
 	/* mutable type */
- 	rep_array = 13,
-	rep_struct = 14,
-	rep_union = 15,
-	rep_hash = 16
+ 	rep_array = 14,
+	rep_struct = 15,
+	rep_union = 16,
+	rep_hash = 17
 } Rep;
 
 /* because rep_undef is -1, using (unsigned) makes these a single compare */
@@ -299,6 +302,7 @@ extern ValueRep    StringRep, ArrayRep, FileRep;
 extern ValueRep    RefRep, StructRep, UnionRep, HashRep;
 extern ValueRep	   FuncRep, ThreadRep;
 extern ValueRep    SemaphoreRep, ContinuationRep, UnitRep, BoolRep;
+extern ValueRep    ForeignRep;
 
 #define NewInt(i)	((Value) IntToPtr ((((i) << 1) | 1)))
 #define IntSign(i)	((i) < 0 ? Negative : Positive)
@@ -353,6 +357,7 @@ extern ValueRep    SemaphoreRep, ContinuationRep, UnitRep, BoolRep;
 #define ValueIsContinuation(v) (ValueRep(v) == &ContinuationRep)
 #define ValueIsUnit(v) (ValueRep(v) == &UnitRep)
 #define ValueIsBool(v) (ValueRep(v) == &BoolRep)
+#define ValueIsForeign(v) (ValueRep(v) == &ForeignRep)
 
 /*
  * Aggregate types
@@ -576,6 +581,14 @@ typedef struct _string {
 
 #define StringChars(s)	    ((char *) ((s) + 1))
 
+typedef struct _foreign {
+    BaseValue	    base;
+    const char	    *id;
+    void	    *data;
+    void	    (*mark)(void *);
+    void	    (*free)(void *);
+} Foreign;
+
 /*
  * Resizable arrays are actually vectors of single entry
  * boxes.  Otherwise shrinking the array leaves old references
@@ -652,6 +665,7 @@ typedef struct _file {
 #define FileEnd		    0x0400
 #define FileString	    0x0800
 #define FilePipe	    0x1000
+#define FileIsPipe	    0x2000
 
 typedef struct _boxTypes {
     DataType	*data;
@@ -793,7 +807,7 @@ void	ContinuationInit (ContinuationPtr continuation);
 
 #ifdef DEBUG_JUMP
 void	    ContinuationTrace (char *where, Continuation *continuation, int indent);
-void	    ContinuationTrace (char	*where, Value continuation);
+void	    ThreadCatches (Value thread);
 #endif
 
 /*
@@ -853,6 +867,7 @@ typedef union _value {
     Array	array;
     File	file;
     Ref		ref;
+    Foreign	foreign;
     Struct	structs;
     Union	unions;
     Func	func;
@@ -950,7 +965,9 @@ Value	NewIntegerFloat (Integer *i, unsigned prec);
 Value	NewNaturalFloat (Sign sign, Natural *n, unsigned prec);
 Value	NewRationalFloat (Rational *r, unsigned prec);
 Value	NewValueFloat (Value av, unsigned prec);
+Value	NewDoubleFloat (double d);
 Value	NewContinuation (ContinuationPtr continuation, InstPtr pc);
+Value	NewForeign (const char *id, void *data, void (*mark)(void *data), void (*free)(void *data));
 
 unsigned    FpartLength (Fpart *a);
 
@@ -960,7 +977,7 @@ unsigned    FpartLength (Fpart *a);
 extern DataCachePtr	refCache;
 
 Value	NewString (long length);
-Value	NewStrString (char *);
+Value	NewStrString (const char *);
 Value	NewCharString (int c);
 Value	NewArray (Bool constant, Bool resizable, TypePtr type, int ndim, int *dims);
 void	ArrayResize (Value av, int dim, int size);
@@ -1010,8 +1027,11 @@ Type	*StructMemType (StructType *st, Atom name);
 Value	StructMemRef (Value sv, Atom name);
 Value	StructMemValue (Value sv, Atom name);
 Value	NewUnion (StructType *type, Bool constant);
+Type	*BuildUnionType (int nelements, ...);
+Type	*BuildEnumType (int nelements, ...);
 Value	UnionValue (Value uv, Atom name);
 Value	UnionRef (Value uv, Atom name);
+Type	*BuildArrayType (Type *type, int ndim, ...);
 
 Value	BinaryOperate (Value av, Value bv, BinaryOp operator);
 Value	UnaryOperate (Value v, UnaryOp operator);
@@ -1068,7 +1088,7 @@ void	FileVPrintf (Value, char *, va_list);
 void	FileSetBuffer (Value file, int buf);
 
 extern Bool	anyFileWriteBlocked;
-extern Bool	anyFileReadBlocked;
+extern Bool	anyPipeReadBlocked;
 
 extern Value    FileStdin, FileStdout, FileStderr;
 
@@ -1106,6 +1126,7 @@ Value	Lxor(Value, Value), Lnot (Value);
 Value   Popcount(Value);
 Bool	Print (Value, Value, char format, int base, int width, int prec, int fill);
 void	PrintError (char *s, ...);
+HashValue HashCrc32 (unsigned char *bytes, int nbytes);
 Value	CopyMutable (Value v);
 #ifdef HAVE_C_INLINE
 static inline Value
@@ -1176,6 +1197,8 @@ int	NaturalToInt (Natural *);
 int	IntegerToInt (Integer *);
 int	IntPart (Value, char *error);
 
+double	DoublePart (Value av, char *error);
+    
 #ifndef Numericp
 Bool	Numericp (Rep);
 #endif
@@ -1204,6 +1227,7 @@ int	FpartInit (void);
 int	StringInit (void);
 int	StructInit (void);
 int	RefInit (void);
+int	ForeignInit (void);
 int	ValueInit (void);
 
 # define oneNp(n)	((n)->length == 1 && NaturalDigits(n)[0] == 1)
